from aws_iam_utils.constants import READ, WRITE, LIST

"""
This is a mapping of actions to overridden action data. This is primarily used to
provide missing/incorrect results from policy_sentry's get_action_data.
"""
ACTION_DATA_OVERRIDES = {
    x["action"].lower(): x
    for x in [
        {
            "action": "lambda:CreateFunctionUrlConfig",
            "access_level": WRITE,
        },
        {
            "action": "lambda:GetFunctionurlConfig",
            "access_level": READ,
        },
        {
            "action": "lambda:DeleteFunctionUrlConfig",
            "access_level": WRITE,
        },
        {
            "action": "lambda:UpdateFunctionUrlConfig",
            "access_level": WRITE,
        },
        {
            "action": "lambda:ListFunctionurlConfigs",
            "access_level": LIST,
        },
        {
            "action": "lambda:InvokeFunctionUrl",
            "access_level": READ,
        },
        {
            "action": "events:ListEndpoints",
            "access_level": LIST,
        },
        {
            "action": "events:DescribeEndpoint",
            "access_level": READ,
        },
        {
            "action": "wafv2:ListMobileSdkReleases",
            "access_level": LIST,
        },
        {
            "action": "wafv2:GetMobileSdkRelease",
            "access_level": READ,
        },
    ]
}



import aws_iam_utils.checks
import aws_iam_utils.generator
import aws_iam_utils.combiner
import aws_iam_utils.util
import aws_iam_utils.simplifier
import aws_iam_utils.action_data_overrides  # noqa: F401
import aws_iam_utils.policy
import aws_iam_utils.policy_permission_item  # noqa: F401



class PolicyPermissionItem:
    def __init__(self, effect, action, resource=None, condition=None, principal=None):
        self.effect = effect
        self.action = action
        self.resource = resource
        self.condition = condition
        self.principal = principal

    def as_statement(self):
        result = {
            "Effect": self.effect,
            "Action": self.action,
        }

        if self.resource is not None:
            result["Resource"] = self.resource
        if self.condition is not None:
            result["Condition"] = self.condition
        if self.principal is not None:
            result["Principal"] = self.principal

        return result

    def __as_dict(self):
        return {
            "effect": self.effect,
            "action": self.action,
            "resource": self.resource,
            "condition": self.condition,
            "principal": self.principal,
        }

    def __repr__(self):
        return f"PolicyPermissionItem({self.__as_dict()})"

    def __eq__(self, other):
        if type(other) is PolicyPermissionItem:
            return self.__as_dict() == other.__as_dict()
        else:
            return self.__as_dict() == other



from aws_iam_utils.util import extract_policy_permission_items
from aws_iam_utils.combiner import collapse_policy_statements
from aws_iam_utils.policy_permission_item import PolicyPermissionItem


def policy_from_dict(policy):
    ppis = [PolicyPermissionItem(**x) for x in extract_policy_permission_items(policy)]

    return Policy(version=policy["Version"], ppis=ppis)


class Policy:
    def __init__(self, version: str, ppis: list[PolicyPermissionItem]):
        self.version = version
        self.ppis = ppis

    def as_dict(self):
        statements = [p.as_statement() for p in self.ppis]

        return collapse_policy_statements(
            {
                "Version": self.version,
                "Statement": statements,
            }
        )

    def find_action_ppis(self, action_name):
        l_action_name = action_name.lower()
        result = []
        for ppi in self.ppis:
            if ppi.action.lower() == l_action_name:
                result.append(ppi)

        return result

    def add_policy_statements(self, policy):
        """Adds statements from the given policy into this policy."""
        self.ppis.extend(policy_from_dict(policy).ppis)



def simplify_policy(p: dict) -> dict:
    """For the given policy, simplify any one-item arrays into straight strings, for
    Actions, Principals and Resources."""

    statements = p["Statement"]

    for statement in statements:
        for k in ["Action", "Resource"]:
            if k in statement:
                if type(statement[k]) is list and len(statement[k]) == 1:
                    statement[k] = statement[k][0]

        if "Principal" in statement:
            principal = statement["Principal"]

            for k in ["AWS", "Service"]:
                if k in principal:
                    if type(principal[k]) is list and len(principal[k]) == 1:
                        principal[k] = principal[k][0]

    return p



from policy_sentry.querying.actions import get_action_data
from aws_iam_utils.action_data_overrides import ACTION_DATA_OVERRIDES


def create_policy(*statements: dict, version: str = "2012-10-17") -> dict:
    """Shortcut function to create a policy with the given statements."""
    return {"Version": version, "Statement": list(statements)}


def statement(
    effect: str = "Allow",
    actions: list[str] = [],
    resource: str = None,
    condition: dict = None,
    principal: dict = None,
) -> dict:
    """Shortcut function to create a Statement with the given properties."""
    st = {}

    for k, v in {
        "Effect": effect,
        "Action": actions,
        "Resource": resource,
        "Principal": principal,
        "Condition": condition,
    }.items():
        if v is not None:
            st[k] = v

    return st


def extract_policy_permission_items(
    policy: dict, allow_unsupported: bool = False
) -> dict:
    """
    For every individual permission granted, we build a list of
    { permission, resource, condition, principal } ("permission items").

    The policy is always expanded via expand_policy() first.

    This is useful for comparisons. Currently it does NOT support
    NotAction, NotPrincipal, NotResource keys. The presences of
    those keys will result in an exception, unless allow_unsupported is True.
    """

    items = []

    policy_expanded = policy

    for statement in policy_expanded["Statement"]:
        if not allow_unsupported:
            for k in ["NotAction", "NotPrincipal", "NotResource"]:
                if k in statement:
                    raise ValueError(
                        f"""Policy key {k} is not supported by
                    extract_policy_permission_items() and will be ignored. To
                    ignore this error, call extract_policy_permission_items
                    with allow_unsupported=True."""
                    )

        for k in ["Action", "Resource"]:
            if type(statement.get(k)) is str:  # turn into list
                statement[k] = [statement[k]]

        effect = statement.get("Effect")
        condition = statement.get("Condition")
        principal = statement.get("Principal")

        for resource in statement.get("Resource", [None]):
            for action in statement["Action"]:
                items.append(
                    {
                        "effect": effect,
                        "action": action.lower(),
                        "resource": resource,
                        "condition": condition,
                        "principal": principal,
                    }
                )

    return items


def dedupe_list(lst: list) -> list:
    return sorted(set(lst), key=lambda x: lst.index(x))


def dedupe_policy(policy: dict) -> dict:
    """Deduplicates all Actions, Principals and Resources in the given
    policy."""
    for statement in policy["Statement"]:
        for k in ["Action", "Resource"]:
            if type(statement.get(k)) is list:
                statement[k] = dedupe_list(statement[k])

        principal = statement.get("Principal")
        for k in ["AWS", "Service"]:
            if type(principal.get(k)) is list:
                principal[k] = dedupe_list(principal[k])

    return policy


def get_action_data_with_overrides(service_name: str, action_name: str) -> dict:
    full_action_name = f"{service_name}:{action_name.lower()}"
    if full_action_name in ACTION_DATA_OVERRIDES:
        return {service_name: [ACTION_DATA_OVERRIDES[full_action_name]]}

    return get_action_data(service_name, action_name)


def lowercase_policy(p):
    """Returns policy p, but with all actions and effects in lowercase, to
    match the behaviour of policyuniverse-expanded policies, so we can easily
    compare policies."""
    new_statements = []

    for statement in p["Statement"]:
        new_statement = {}

        # do NOT do Effect lower(), as this breaks policyuniverse
        new_statement["Effect"] = statement["Effect"]

        for key in ["Action"]:
            if key in statement:
                if type(statement[key]) is list:
                    new_statement[key] = [s.lower() for s in statement[key]]
                else:
                    new_statement[key] = statement[key].lower()

        new_statements.append(new_statement)

        for key in ["Principal", "Condition", "Resource"]:
            if key in statement:
                new_statement[key] = statement[key]

    return {
        "Version": p["Version"],
        "Statement": new_statements,
    }


def create_lowercase_policy(*st):
    """Simply wraps create_policy and lowercase_policy, creating a policy with
    lowercase actions and effects regardless of the inputs."""
    return lowercase_policy(create_policy(*st))
